"""
@file
@brief Shortcuts to *search_rank*.
"""

from .search_engine_predictions import SearchEnginePredictions
from .search_engine_predictions_images import SearchEnginePredictionImages
from .search_engine_vectors import SearchEngineVectors
